#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from glob import glob
import os
from rmsd import kabsch_rmsd

def psi4_out_read(out:str):
    """
    Parameters
    ----------
    out : str
        psi4 put file name.
    Returns
    -------
    check_end, check_freq, Electronic energy, Gibbs free energy, exyz, xyz2RMSD_H.
    """

    with open(out, 'rt', encoding='latin-1') as file:
        lines = file.readlines()

    if lines[-1].strip() == '*** Psi4 exiting successfully. Buy a developer a beer!':
        check_end = True
    else:
        check_end  = False
    check_freq = False
    E = 0
    G = 0
    exyz = []
    xyz2RMSD_H =[]

    for i, _ in enumerate(lines):
        if 'Final (previous) structure:' in lines[i]:
            E = float(lines[i-1].split()[-1])
            for j in range(i+2,len(lines)):
                if 'Saving final (previous) structure.' in lines[j]: break
                split = lines[j].split()
                exyz.append([split[0], float(split[1]), float(split[2]), float(split[3])])
                if 'H' not in lines[j]:
                    xyz2RMSD_H.append([float(split[1]), float(split[2]), float(split[3])])
        if 'Freq [cm^-1]' in lines[i]:
            check_freq = True
            freqs = lines[i].split()[2:]
            for f in freqs:
                # psi4 represent the imaginary number as 5i,
                # if an error ocurreduring the float conversion, 
                # or it was converted but if less than cero (another error on sqrt) 
                # then check_freq = False
                try:
                    np.sqrt(float(f))
                except:
                    check_freq = False
                    break
        if 'Total G, Free enthalpy at  298.15 [K]' in lines[i]:
            G = float(lines[i].split('[K]')[1].split()[0])
        
    exyz = pd.DataFrame(exyz)
    xyz2RMSD_H = np.array(xyz2RMSD_H, dtype=float)
    return check_end, check_freq, E, G, exyz, xyz2RMSD_H

def main(SubDirs:bool = True, engine:str = 'psi4', xyz_out:bool = False) -> dict:
    """Process the output of the QM calculations generated through :meth:`aleimi.extractor.main`.
    This function is quiet limited for now, Only useful for psi4 engine.

    Parameters
    ----------
    SubDirs : bool, optional
        Should be True if :meth:`aleimi.extractor.main` was used with ``mkdir = True``, by default True
    engine : str, optional
        psi4, gaussian or orca. It depends on the engine defined on :meth:`aleimi.extractor.main` was used with ``engine`` keyword, by default 'psi4'
    xyz_out : bool, optional
        If True, it will write the xyz coordinates of the conformer with the lowest energy, by default False

    Returns
    -------
    dict
        conf_name: 3D_coordinates

    Raises
    ------
    FileNotFoundError
        If there is not .out files
    """
    if SubDirs:
        outs = [out for out in glob('*/*.out') if 'myjob' not in out]

        first_names = set()
        for out in outs:
            first_names.add(os.path.basename(out).split('_Cell')[0])
        first_names = list(first_names)
    else:
        outs = [out for out in glob('*.out') if 'myjob' not in out]
        
        first_names = set()
        for out in outs:
            first_names.add(out.split('_Cell')[0])
        first_names = list(first_names)
    if not outs: raise FileNotFoundError('No se pueden encontrar los archivos .out')

    if engine == 'psi4':
        reader = psi4_out_read
    elif engine == 'orca':
        pass
    elif engine == 'gaussian':
        pass
    else:
        pass

    wrong_freq = []
    wrong_end = []

    coord_lower_energy = {}
    for first_name in first_names:
        to_work = []
        if SubDirs:
            for out in outs:
                if os.path.basename(out).split('_Cell')[0] == first_name:
                    to_work.append(out)
        else:
            for out in outs:
                if out.split('_Cell')[0] == first_name:
                    to_work.append(out)

        Gibbs_free_energies = []
        coords = []
        coord2RMSD_Hs =[]
        good = []

        for item in to_work:
            check_end, check_freq, E_total_energy, Gibbs_free_energy, coord, coord2RMSD_H = reader(item)
            if check_end:
                if check_freq==True:
                    good.append(os.path.basename(item))
                    Gibbs_free_energies.append(Gibbs_free_energy)
                    coords.append(coord)
                    coord2RMSD_Hs.append(coord2RMSD_H)
                else:
                    wrong_freq.append(item)
            else:
                wrong_end.append(item)
        if good:
            paired = list(zip(good, Gibbs_free_energies, coords, coord2RMSD_Hs)) 
            ORDERED = sorted(paired, key=lambda x: x[1])
            coord_lower_energy[ORDERED[0][0].split('.')[0]] = ORDERED[0][2]
            to_print = []
            for i, _ in enumerate(ORDERED):
                to_print.append([ORDERED[i][0].split('_Cell_')[-1].split('.')[0], ORDERED[i][1]])
            to_print = pd.DataFrame(to_print, columns = ['Cell', 'Gibbs (hartree/partícula)'])    
            print(first_name)
            print(to_print)
            print('\n')
            with open(first_name+'.orden', 'wt') as final:
                to_print.to_string(final)

        
            rmsd_matrix = np.zeros((len(ORDERED),len(ORDERED)))
            for i, _ in enumerate(rmsd_matrix):
                for j, _ in enumerate(rmsd_matrix):
                    if j < i:
                        rmsd_matrix[i][j] = rmsd_matrix[j][i]
                    elif j == i:
                        rmsd_matrix[i][j] == 0.0
                    else:
                        # =============================================================
                        #     Calculando RMSD
                        # =============================================================
                        P = ORDERED[i][3]
                        Q = ORDERED[j][3]
                        rmsd_matrix[i][j] = kabsch_rmsd(P, Q, translate = True)
            index = [g.split('_Cell_')[-1].split('.')[0] for g in good]
            to_print_rmsd_matrix = pd.DataFrame(rmsd_matrix, index = index, columns = index)
            with open(first_name+'.rmsd', 'wt') as final:
                to_print_rmsd_matrix.to_string(final)
        #========================Se exporta el .xyz==================
            if xyz_out:
                with open(ORDERED[0][0].split('.')[0]+'.xyz', 'wt') as final:
                    final.write(str(len(ORDERED[0][2]))+'\n\n')
                    ORDERED[0][2].to_string(final, header=False, index=False)

        
    #======================================================================
    if wrong_freq:
        print(f'Las cálculos: {wrong_freq} presentaron frecuencias negativas o imaginarias.')
    if wrong_end:
        print(f'Las cálculos: {wrong_end} presentaron problemas para un correcta finalizacion. Check the psi4 output!')

    return coord_lower_energy


if __name__ == '__main__':...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# TODO
# Toma de un archivo .smi los SMILES, genera la cant de conformeros seleccionados
# e imprime tanto los .sdf con todos los conformeros como .mop para optimizar en MOPAC
# con el nivel de teoria seleccionado.


from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Draw
import pandas as pd
import os
from typing import List

#      CHECKING geometry degeneracy
# La primera iteracion que corre por i busca que este optimizada la estructura
# La segunda busca que no se geometricamente la misma estructura
#Pueda suceder que alguna idx cumpla que no es geometricamente la misma estructura
#y no la ve el segundo ciclo
#pero el primero cuando llegue a ella la vera y tomara su id
# =============================================================================
def confgen(mol:Chem.rdchem.Mol, rdkit_d_RMSD:float, numConfs:int, rdkit_numThreads:float = 0, UFF:bool = False):
    """Generate conformers stochastically for mol and filter them out based on geometrical and energetic criterions.

    Parameters
    ----------
    mol : Chem.rdchem.Mol
        The RDKit molecule for which the stochastic conformations will be generated.
    rdkit_d_RMSD : float
        RMSD to filter out redundant conformations, geometric filter.
    numConfs : int
        Maximum number of conformation to generate
    rdkit_numThreads : float, optional
        Number of CPUs to use, by default 0 (use all)
    UFF : bool, optional
        If True a classical mechanical optimization will be performed, by default False

    Returns
    -------
    tuple
        * RDKit molecule with conformations, the hydrogens are always added.
        * Index of conformer to use.
        * List of classic energies if UFF = True, if not, a list of None values with length, number of conformers.
    """
    molecule = Chem.AddHs(mol)
    ids = AllChem.EmbedMultipleConfs(molecule, numConfs=numConfs, numThreads=rdkit_numThreads)
    if UFF:
        opt = AllChem.UFFOptimizeMoleculeConfs(molecule, numThreads=rdkit_numThreads, ignoreInterfragInteractions=False, maxIters=500) # The result is a list a containing 2-tuples: (not_converged, energy) for each conformer. If not_converged is 0, the minimization for that conformer converged.

        print('Calculando RMSD ...')
        rmsd_reject = set()
        opt_reject = set()
        rmsd_no_reject= set()
        no_opt=0
        ids=list(ids)
        for i in ids:

            if opt[i][0]==1:
                no_opt +=1
                opt_reject.add(ids[i])
            else:
                for idx in ids:
                    if i < idx:
                        RMSD = AllChem.GetConformerRMS(molecule, i, idx, prealigned=True)
                        if RMSD <= rdkit_d_RMSD:
                            if opt[i][1] < opt[idx][1]:
                                rmsd_reject.add(ids[i])
                                rmsd_no_reject.add(ids[idx])
                            else:
                                rmsd_reject.add(ids[idx])
                                rmsd_no_reject.add(ids[i])

        reject = opt_reject.union(rmsd_reject)
        if len(rmsd_reject) !=0:
            print(f'Los confórmeros con Id {rmsd_reject} no se imprimirán en el .mop por ser iguales (Según el error de RMSD = {rdkit_d_RMSD}) a {rmsd_no_reject} respectivamente y menos favorables energéticamente que los últimos mencionados.')
        if len(opt_reject) !=0:
            print(f'Los confórmeros con Id {opt_reject} no se imprimirán en el .mop pues no lograron ser optimizados.')
        to_use = list(set(ids) - reject)
        return molecule, to_use, [item[1] for item in opt]# Here I am creating a list of tuple (molecule, energy)
    else:

        print('Calculando RMSD ...')
        rmsd_reject = set()
        rmsd_no_reject= set()
        ids=list(ids)
        for i in ids:
            for idx in ids:
                if i < idx:
                    RMSD = AllChem.GetConformerRMS(molecule, i, idx, prealigned=True)
                    if RMSD <= rdkit_d_RMSD:
                        rmsd_reject.add(ids[i])
                        rmsd_no_reject.add(ids[idx])


        if len(rmsd_reject) !=0:
            print(f'Los confórmeros con Id {rmsd_reject} no se imprimirán en el .mop por ser iguales (Según el error de RMSD = {rdkit_d_RMSD}) a {rmsd_no_reject} respectivamente y menos favorables energéticamente que los últimos mencionados.')

        to_use = list(set(ids) - rmsd_reject)
        return molecule, to_use, len(ids)*[None] # Because I did not optimize, then I add None

# =============================================================================
#exportar las imagenes de la mejor manera posible. en columnas de 1,2,3,4 o 5
# =============================================================================

def makeimg(mols:List[Chem.rdchem.Mol], **keywords):
    """Simple function to create a picture of the input molecules.
    The file ``used_mols.png`` will be createad.

    Parameters
    ----------
    mols : List[Chem.rdchem.Mol]
        list of molecules
    """

    ms = [mol for mol in mols if mol is not None]
    _ = [AllChem.Compute2DCoords(m) for m in ms]

    cant = len(ms)
    if cant==1:
        best=1
    else:
        div = range(2,6)
        tup = [[i - cant%i,i] for i in div]

        for t in tup:
            if t[0] == t[1]:
                t[0] = 0

        minimum_rest = min(tup, key=lambda x: x[0])
        for t in tup:
            if t[0] == minimum_rest[0] and t[1]>minimum_rest[1]:
                minimum_rest = t
        best = minimum_rest[1]

    if 'legends' in keywords:
        legends = keywords['legends']
    else:
        legends = [f'Mol: {x+1}' for x in range(len(ms))]

    img=Draw.MolsToGridImage(ms,molsPerRow=best,subImgSize=(700,700), legends=legends)
    if len(ms) > 1:
        img.save('used_mols.png')
    else:
        img.save(f'{legends[0]}.png')



def main(suppl:str, numConfs:int = 10, rdkit_d_RMSD:float = 0.2, UFF:bool = False, rdkit_numThreads:int = 0, mopac_keywords:str =  'PM7 precise ef xyz geo-ok t=3h THREADS = 2') -> List[str]: #EPS=78.4
    """Generate conformers stochastically for the supplied molecules and filter them out based on geometrical and energetic criterions.

    Parameters
    ----------
    suppl : str
        Path for the molecule file. It could be:.pdb, .mol2, .mol, .sdf, .smi. In case of .smi you can define several SMILES.
        and for all of them the configuration will be generated.
        TODO: do the same for sdf files
    numConfs : int, optional
        Maximum number of conformation to generate, by default 10
    rdkit_d_RMSD : float, optional
        RMSD to filter out redundant conformations, geometric filter, by default 0.2
    UFF : bool, optional
        If True a classical mechanical optimization will be performed, by default False
    rdkit_numThreads : int, optional
        Number of CPUs to use, by default 0 (use all)
    mopac_keywords : str, optional
        This is head of a .mop file, basically a definition of the sei-empirical job. Check
        the `MOPAC Documentation <http://openmopac.net/>`_ for more information , by default 'PM7 precise ef xyz geo-ok t=3h THREADS = 2'

    Returns
    -------
    List[str]
        List of names of the used molecules

    Raises
    ------
    ValueError
        If in the .smi there are invalid SMILES
    ValueError
        If the extension of molecule is not: .pdb, .mol2, .mol, .sdf or .smi.
    ValueError
        If the molecule was not understood by RDKit
    """

    name, ext = os.path.splitext(os.path.basename(suppl))
    print('Resumen\n')
    if ext == ".smi":
        with open(suppl, 'rt') as file:
            smiles = file.readlines()

        # =============================================================================
        #    Comprobamos que los SMILES esten bien
        # expuestas con la funcion molfilter y exportamos .sdf
        # =============================================================================
        Errors_mols = []
        for i, item in enumerate(smiles):
            try:
                Chem.MolFromSmiles(item).GetNumAtoms()
            except:
                Errors_mols.append(i+1)
        if len(Errors_mols) !=  0:
            raise ValueError(f'Las moléculas con Ids: {Errors_mols} no son estructuras SMILES correctas para RDKit. Por favor, retirelas del .smi.')


        mols = [(f'conf_{name}_{i+1}', Chem.MolFromSmiles(smile)) for (i,smile) in enumerate(smiles)]
    elif ext == '.pdb':
        mols = [(f"conf_{name}", Chem.MolFromPDBFile(suppl))]
    elif ext == '.mol2':
        mols = [(f"conf_{name}", Chem.MolFromMol2File(suppl))]
    # elif ext == '.mol':
    #     mols = [(f"conf_{name}", Chem.MolFromMolFile(suppl))]
    elif ext in ['.mol', '.sdf']:
        mols = [(f"conf_{name}", Chem.MolFromMolFile(suppl))]
    else:
        raise ValueError(f'The structure must be one of the following extensions: .pdb, .mol2, .mol, .sdf, .smi. {suppl} was provided')

    if None in [mol[1] for mol in mols]:
        raise ValueError(f"{suppl} is not understanded by RDKit")




    # =============================================================================
    # Generamos conformaciones que cumplen con las condiciones de rmsd
    # expuestas con la funcion molfilter y exportamos .sdf
    # =============================================================================

    makeimg([mol[1] for mol in mols], legends = [mol[0] for mol in mols])
    for (i, mol) in enumerate(mols):

        print(f'Se están generando {numConfs} conformaciones para la molécula con Id = {i+1}...')

        molecule, index, opt = confgen(mol[1], rdkit_d_RMSD, numConfs, rdkit_numThreads = rdkit_numThreads, UFF = UFF)
        natoms = molecule.GetNumAtoms()

        sdf_writer = Chem.SDWriter(f"{mol[0]}.sdf")
        for idx in index:
            sdf_writer.write(molecule, confId=idx)
        sdf_writer.close()
        print(f'{mol[0]}:  {len(index)}/{numConfs}')

    # =============================================================================
    #     Generando los .mop
    # =============================================================================


        print(f"Archivos de salida: {mol[0]}.sdf, {mol[0]}.mop")

        # TODO Change this part to work properlly with the attribute _Name of the molecule
        with open(f"{mol[0]}.sdf", 'rt') as file:
            lines = file.readlines()

        final=open(f"{mol[0]}.mop",'w')

        #Toma los valores reales que se usaron para imprimir el .sdf y asi buscar los opt de verdad, pq si se elimino una estructura tengo que tenerlo en cuenta
        cont = 0 #Lo inicializao en -1 pq la cantidad de veces que se encontrara la palabra RDKit coincide con la cantidad con el len(to_use[0])
        #todo esto para que en el .mop me saque el valor de energia optenido en la optimizacion
        for k, line in enumerate(lines):
            if 'RDKit          3D' in line:

                chunk = lines[(k+3):(k+3+(natoms))]
                sliced = []
                for c in chunk:
                    sliced.append(c.split())
                df = pd.DataFrame(sliced)
                to_print = df[[3, 0, 1, 2]]
                final.write(mopac_keywords+'\n')
                try:
                    comments = f"{mol[0]}   E_UFF = {round(float(opt[index[cont]]), 3)}"
                except:
                    comments = f"{mol[0]}   E_UFF = {opt[index[cont]]}"
                final.write(comments+'\n')
                final.write('CELL: %d\n' % (cont+1))
                to_print.to_string(final, header=False, index=False)
                final.write('\n0\n')
                cont += 1
        print('Número de átomos: %d\n'% (natoms))
        final.close()
    mol_names = [mol[0] for mol in mols]
    return mol_names


if __name__ == '__main__':...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pandas as pd
import os
from aleimi import templates, utils
from typing import List

#def extract(arc_file, boltzmann_file, energy_cut = 2, conformer_cut = None, mksh = True, mkdir = True):
# Energy cut is in kcal/mol
def extract(boltzmann_file:str, energy_cut:float = 2.0, conformer_cut:int = None) -> List[int]:
    """Extract the conformers based on the filters: energy_cut and/or conformer_cut

    Parameters
    ----------
    boltzmann_file : str
        The boltzmann file generated with `:meth: aleimi.boltzmann.main`
    energy_cut : float, optional
        Maximum difference in energy with respect the conformer with lowest energy, by default 2.0
    conformer_cut : int, optional
        Maximum number of conformers to export, by default None

    Returns
    -------
    List[int]
        The list of ``cell`` identifiers
    """

    df = pd.read_csv(boltzmann_file)
    indx_to_extract = []
    if energy_cut:
        df_subset = df[df.Emin_Ei >= -energy_cut]
        indx_to_extract += df_subset.cell.tolist()

    if conformer_cut:
        df_subset = df.iloc[:conformer_cut]
        indx_to_extract += df_subset.cell.tolist()

    if not energy_cut and not conformer_cut:
        df_subset = df.iloc[:]
        indx_to_extract += df_subset.cell.tolist()

    return sorted(indx_to_extract)

def get_coords(input_file, indx_to_extract):
    
    file_ext = os.path.basename(input_file).split('.')[-1]
    file_name = os.path.basename(input_file)[:-(len(file_ext)+1)]

    if file_ext == 'arc':
        with open(input_file, 'rt', encoding='latin-1') as file:
            lines = file.readlines()
        
        to_return = []
        for line in lines:
            if 'Empirical Formula' in line:
                natoms = int(line.split()[-2])
                break
        for i, line in enumerate(lines):
            if ('FINAL GEOMETRY OBTAINED' in line) and (int(lines[i+3].strip().split(':')[1]) in indx_to_extract):
                cell = int(lines[i+3].strip().split(':')[1])
                chunk = lines[i+4:i+4+natoms]
                sliced = []
                for c in chunk:
                    sliced.append(c.split())
                df = pd.DataFrame(sliced)
                # Creating a tuple (conf_name_Cell, coords)
                to_return.append((f"{file_name}_Cell_{cell}", df[[0, 1, 3, 5]]))
        return to_return

    elif file_ext == 'out':
        f = open(input_file, 'r')
        chunk = []
        cart = []


        # getting data from out
        # finding No. of atoms
        
        while True:
            line = f.readline()
            if "Empirical Formula" in line:
                natoms = int(line.split()[-2])
                f.close()
                break
                            
        f = open(input_file, 'r')
        to_return = []
        while True:
            line = f.readline()
            if len(line) == 0:break
                
            if 79*'-' in line:
                while True:
                    line = f.readline()
                    if (79*'*' in line) or (len(line) == 0):break

                    if 'CELL' in line: 
                        cell = int(line.split(':')[1])

                    elif 'CARTESIAN COORDINATES' in line and cell in indx_to_extract:
                        utils.ignoreLines(f, 1)
                        cont = 0
                        chunk = []        
                        while cont < natoms:
                            chunk.append(f.readline())
                            cont += 1
                        cart = []
                        for c in chunk:
                            cart.append(c.split())
                            df = pd.DataFrame(cart)
                            to_return.append((f"{file_name}_Cell_{cell}",df[[1, 2, 3, 4]]))  
        return to_return
    else:
        raise ValueError(f"{input_file} does not have .arc or .out extension. Therefore is not readeable by ALEIMI.")

def main(
    input_file,
    boltzmann_file,
    energy_cut = 2,
    conformer_cut = 0,
    engine = 'psi4',
    machine = 'smaug',
    mkdir = True,
    jobsh = True,
    **keywords):
    
    if engine == 'psi4':
        InputExt = '.in'
    elif engine == 'orca':
        InputExt = '.inp'
    elif engine == 'gaussian':
        InputExt = '.gjf'
    else:
        print(f"Warning!: It was used 'in' as generic extension for the input file for the non recognized engine: {engine}")
        InputExt = '.in'
    
    indx_to_extract = extract(boltzmann_file, energy_cut = energy_cut, conformer_cut = conformer_cut)
    names_coords = get_coords(input_file, indx_to_extract)
    for name, coords in names_coords:
        INPUT_obj = templates.INPUT(engine, machine = machine, name = name, coords = coords, **keywords)
        if mkdir:
            utils.makedirs(name)
            INPUT_obj.write(os.path.join(name, f"{name}{InputExt}"),'input')
            if jobsh:
                INPUT_obj.write(os.path.join(name, f"job.sh"),'jobsh')
        else:
            INPUT_obj.write(f"{name}{InputExt}",'input')
            if jobsh:
                INPUT_obj.write(f"{name}.sh",'jobsh')

if __name__ == '__main__':...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from aleimi import confgen, boltzmann, extractor, utils
import os, yaml, tempfile

cwd = os.getcwd()
tmp_path = tempfile.TemporaryDirectory(prefix='aleimi_wd_')
os.chdir(tmp_path.name)
suppl1 = 'suppl1.smi'
suppl2 = 'suppl2.smi'

with open(suppl1, 'w') as s:
    s.write("COC(=O)C=1C=CC(=CC1)S(=O)(=O)N\nCCO")
with open(suppl2, 'w') as s:
    s.write("COC(=O)C\nCCO\nCNCO")

def test_module():

    mol_names = confgen.main('suppl1.smi')
    for mol_name in mol_names:
        print(mol_name)
        utils.mopac(f"{mol_name}.mop")
        boltzmann.main(f"{mol_name}.arc")
        extractor.main(f"{mol_name}.arc",f"{mol_name}_boltzmann.csv")

def test_cli_default_kwargs():
    utils.run('aleimi-run suppl2.smi')

def test_cli_user_kwargs():
    # Save the config as a yaml file
    params = {
    # aleimi.confgen.main
    'numConfs': 5,
    'rdkit_d_RMSD': 0.1,
    'UFF': True,
    'rdkit_numThreads': 0,
    'mopac_keywords': 'PM6 precise ef xyz geo-ok t=3h',

    # aleimi.boltzmann.main
    'Bd_rmsd': 2.0,
    'Bd_E': 0.0001,
    'BOutPath': True,
    
    # aleimi.extractor.main
    "energy_cut": 3,
    "conformer_cut": 4,
    "engine": 'psi4',
    }
    with open('params.yml', 'w') as p:
        yaml.dump(params, p)
    utils.run('aleimi-run suppl2.smi --params params.yml')

if __name__ == '__main__':
    test_module()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from aleimi import confgen, boltzmann, extractor, __version__, utils
import argparse, os, yaml

"""
Tengo que adicionar la parte de los comandos extras para pasarselos a exrtractor
para la creacion de los templates, probar con los argumentos que se pasan
si alguno esta mal entonces dlanzar un warning o algo por el estilo
Tengo que ver esto bien, me falta por implemnetar lo fde .gjf
Y lo de el analisis para al menos psi4 y orca
    """

def _aleimi():
    """CLI of ``aleimi``
    """
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        help = "The path to the directory were the molecule(s) is(are)",
        dest='suppl',
        type=str)
    parser.add_argument(
        '-p', '--params',
        help = "Parameters to run ALEIMI",
        default=None,
        dest='params',
        type=str)
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=f"aleimi: {__version__}")

    args = parser.parse_args()

    suppl = args.suppl
    params = args.params
    if not os.path.exists(args.suppl):
        raise FileNotFoundError(f"{args.suppl} does not exist or is not accessible.")

    # Getting the default kwargs of all the functions.
    confgen_keywords = utils.get_default_kwargs(confgen.main)
    boltzmann_keywords = utils.get_default_kwargs(boltzmann.main)
    extractor_keywords = utils.get_default_kwargs(extractor.main)
    used_keywords = {**confgen_keywords, **boltzmann_keywords, **boltzmann_keywords}
    if args.params:
        if not os.path.exists(params):
            raise FileNotFoundError(f"{params} does not exist or is not accessible.")
        
        with open(args.params, 'r') as params:
            user_keywords =  yaml.safe_load(params)

        for key in confgen_keywords.keys():
            if key in user_keywords:
                try:
                    confgen_keywords[key] = type(confgen_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
                except:
                    raise ValueError(f"{user_keywords[key]} must be a {type(confgen_keywords[key])}-like")

        for key in boltzmann_keywords.keys():
            if key in user_keywords:
                try:
                    boltzmann_keywords[key] = type(boltzmann_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
                except:
                    raise ValueError(f"{user_keywords[key]} must be a {type(boltzmann_keywords[key])}-like")

        for key in extractor_keywords.keys():
            if key in user_keywords:
                try:
                    extractor_keywords[key] = type(extractor_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
                except:
                    raise ValueError(f"{user_keywords[key]} must be a {type(extractor_keywords[key])}-like")

    # Save the config as a yaml file
    with open('outparams.yml', 'w') as f:
        yaml.dump(used_keywords, f)


    mol_names = confgen.main(suppl,**confgen_keywords)
    for mol_name in mol_names:
        print(mol_name)
        utils.mopac(f"{mol_name}.mop")
        boltzmann.main(f"{mol_name}.arc",**boltzmann_keywords)
        extractor.main(f"{mol_name}.arc",f"{mol_name}_boltzmann.csv", **boltzmann_keywords)


if __name__ == '__main__':...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pandas as pd
import copy, os
class INPUT:
    def __init__(self, engine, machine = 'smaug', **keywords):
        self.engine = engine
        self.machine = machine
        self.default_partition = {'smaug': 'deflt', 'gwdg': 'medium'}
        self.default_exclude = {'smaug': "fang[1,11-50]", 'gwdg': None}
        self.default_ntasks = {'smaug': 12, 'gwdg': 24}
        self.default_keywords = {
            'psi4':{
                #Parameters for input
                'theory': 'MP2',
                'basis': '6-31+G*',
                'memory': '10 GB',
                'name': 'mol',
                'charge': 0,
                'multiplicity': 1,
                'PCM': False,
                'coords': pd.DataFrame(),
                'calc_type': 'optimize',
                'freq': True,
                #Parameters for jobsh
                'time': '2-00:00',
                'ntasks': self.default_ntasks[self.machine],
                'cpus-per-task': 1,
                'partition': self.default_partition[self.machine],
                'nodes': 1,
                'nice': 0,
                'gpus': 0,
                'mail_user': None,
                'exclude': self.default_exclude[self.machine],
                },
            'orca':{
                'theory': 'RI-MP2',
                'basis': '6-31+G* def2-TZVP/C',
                'name': 'mol',
                'charge': 0,
                'multiplicity': 1,
                'SMD': False,
                'coords': pd.DataFrame(),
                'calc_type': 'Opt',
                'freq_type': 'NumFreq',
                'SCF_details': 'VeryTightSCF VeryTightOpt',
                'maxcore':2400,
                #Parameters for jobsh
                'time': '2-00:00',
                'ntasks': 6,
                'cpus-per-task': 2,
                'partition': 'deflt',
                'nodes': 1,
                'nice': 0,
                'gpus': 0,
                'mail_user': None,
                'exclude': "fang[1,11-50]" #None
                },
            'gaussian':{
                'theory': 7,
                'basis': 2,
                'memory':4,
                'name':5,
                'charge':0,
                'multiplicity':1,
                'PCM': False,
                'coords': None,
                'calc_type':5,
                'freq': False}
        }
        self.keywords = copy.deepcopy(self.default_keywords)
        for key in keywords:
            self.keywords[engine][key] = keywords[key]
        self.input = ''
        self.jobsh = ''
        self.parse()

    def parse(self):
        if self.engine == 'psi4':
            self.input += f"#! Computation at {self.keywords[self.engine]['theory']}/{self.keywords[self.engine]['basis']} for {self.keywords[self.engine]['name']} \n\n"\
            f"memory {self.keywords[self.engine]['memory']}\n\n"\
            f"molecule {self.keywords[self.engine]['name']} {'{'}\n"\
            f"{self.keywords[self.engine]['charge']} {self.keywords[self.engine]['multiplicity']}\n"\
            f"{self.keywords[self.engine]['coords'].to_string(header=False, index=False)}\n"
            if self.keywords[self.engine]['PCM']: self.input += "symmetry c1\n"
            self.input += "}\n\n"
            if self.keywords[self.engine]['PCM']:
                self.input += f"set {'{'}\n  basis {self.keywords[self.engine]['basis']}\n"\
                "   scf_type pk\n"\
                "   pcm true\n"\
                "   pcm_scf_type total\n"\
                "   }\n\n"\
                "pcm = {\n"\
                "   Units = Angstrom\n"\
                "   Medium {\n"\
                "       SolverType = IEFPCM\n"\
                "       Solvent = Water\n"\
                "       }\n\n"\
                "   Cavity {\n"\
                "       RadiiSet = UFF\n"\
                "       Type = GePol\n"\
                "       Scaling = False\n"\
                "       Area = 0.3\n"\
                "       Mode = Implicit\n"\
                "       }\n"\
                "   }\n"
            else:
                self.input += f"set basis {self.keywords[self.engine]['basis']}\n"
            self.input += f"{self.keywords[self.engine]['calc_type']}('{self.keywords[self.engine]['theory']}')\n"
            if self.keywords[self.engine]['freq']:
               self.input += f"{self.keywords[self.engine]['theory']}_e, {self.keywords[self.engine]['theory']}_wfn = frequencies('{self.keywords[self.engine]['theory']}', return_wfn=True, dertype=1)\n"
            self.input += f"with open('{self.keywords[self.engine]['name']}_psi4out.xyz', 'w') as f:\n"\
            f"\tf.write('{len(self.keywords[self.engine]['coords'])} ' )\n"\
            f"\tf.write('%.12f\\n' % {self.keywords[self.engine]['theory']}_e)\n"\
            f"\tf.write({self.keywords[self.engine]['name']}.save_string_xyz())"

            self.jobsh += "#!/bin/bash\n"\
            f"#SBATCH --partition {self.keywords[self.engine]['partition']}\n"\
            f"#SBATCH --output myjob_{self.keywords[self.engine]['name']}.out\n"\
            f"#SBATCH --error myjob_{self.keywords[self.engine]['name']}.err\n"\
            f"#SBATCH --ntasks {self.keywords[self.engine]['ntasks']}\n"\
            f"#SBATCH --cpus-per-task {self.keywords[self.engine]['cpus-per-task']}\n"\
            f"#SBATCH --time {self.keywords[self.engine]['time']}\n"\
            f"#SBATCH --job-name={self.keywords[self.engine]['name']}\n"\
            f"#SBATCH --nodes {self.keywords[self.engine]['nodes']}\n"\
            f"#SBATCH --nice={self.keywords[self.engine]['nice']}\n"\
            f"#SBATCH --gpus={self.keywords[self.engine]['gpus']}\n"
            if self.keywords[self.engine]['mail_user']:
                self.jobsh += f"#SBATCH --mail-user={self.keywords[self.engine]['mail_user']}\n"
            if self.keywords[self.engine]['exclude']:
                self.jobsh +=f"#SBATCH --exclude={self.keywords[self.engine]['exclude']}\n"

            if self.machine == 'gwdg':
                self.jobsh += f"#SBATCH -A all\n"\
                "#SBATCH -C scratch\n"

            self.jobsh +="\n# This block is for the execution of the program\n"
            if self.machine == 'smaug':
                self.jobsh += "source /home/users/all-jh/opt/miniconda3/etc/profile.d/conda.sh #It is needed to use the conda activate command\n"\
                "conda activate htmd\n\n"\
                "# Creating local scratch folder for the user on the computing node.\n"\
                "MY_TEMP_DIR=\"$(mktemp -d /localdisk/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"
            elif self.machine == 'gwdg':
                self.jobsh += "module load anaconda3/2020.07\n"\
                "source /opt/sw/rev/20.12/haswell/gcc-9.3.0/anaconda3-2020.07-slbv7z/etc/profile.d/conda.sh #It is needed to use the conda activate command\n"\
                "conda activate htmd\n\n"\
                "# Creating local scratch folder for the user on the computing node.\n"\
                "MY_TEMP_DIR=\"$(mktemp -d /scratch/users/${USER}/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"
            else:
                pass

            self.jobsh += "# make sure the temp dir was created\n"\
            "[ -d $MY_TEMP_DIR ] || exit 1\n"\
            "# make sure to delete the directory on (erroneous) exit\n"\
            "trap 'rm -rf -- \"$MY_TEMP_DIR\"' EXIT\n"\
            "# export PSI_SCRATCH env variable\n"\
            "export PSI_SCRATCH=$MY_TEMP_DIR\n\n"\
            "# This block is echoing some SLURM variables\n"\
            "echo \"Job execution start: $(date)\"\n"\
            "echo \"Jobid = $SLURM_JOBID\"\n"\
            "echo \"Host = $SLURM_JOB_NODELIST\"\n"\
            "echo \"Jobname = $SLURM_JOB_NAME\"\n"\
            "echo \"Subcwd = $SLURM_SUBMIT_DIR\"\n"\
            "echo \"SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK\"\n"\
            "echo \"SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE\"\n"\
            "echo \"The temp file used was: $MY_TEMP_DIR\"\n\n"\
            f"psi4 -i ${{SLURM_JOB_NAME}}.in -o ${{SLURM_JOB_NAME}}.out -n {self.keywords[self.engine]['ntasks'] * self.keywords[self.engine]['cpus-per-task']} #run command, in, out and number of threads to be used\n\n"\
            "#Deliting the scratch file. Normal delete on normal exit\n"\
            "rm -rf $MY_TEMP_DIR\n\n"\
            "echo \"Job execution end: $(date)\""

        elif self.engine == 'orca':
            #!! This is not implemented for GWDG
            self.input += f"#! Computation at {self.keywords[self.engine]['theory']}/{self.keywords[self.engine]['basis']} for {self.keywords[self.engine]['name']}\n"\
            f"%pal nprocs {self.keywords[self.engine]['ntasks']} end\n"
            if self.keywords[self.engine]['freq_type']:
                self.input += f"%maxcore {self.keywords[self.engine]['maxcore']}\n\n"\
                f"! {self.keywords[self.engine]['SCF_details']}\n"\
                f"! {self.keywords[self.engine]['freq_type']}\n"
            self.input += f"! {self.keywords[self.engine]['theory']} {self.keywords[self.engine]['basis']}\n"\
            f"! {self.keywords[self.engine]['calc_type']}\n"
            if self.keywords[self.engine]['SMD']:
                self.input += "%cpcm\n"\
                "   smd true\n"\
                "   SMDsolvent \"water\"\n"\
                "end\n"
            self.input += f"\n* xyz {self.keywords[self.engine]['charge']} {self.keywords[self.engine]['multiplicity']}\n"\
            f"! {self.keywords[self.engine]['coords'].to_string(header=False, index=False)}\n*"

            self.jobsh += "#!/bin/bash\n"\
            f"#SBATCH --partition {self.keywords[self.engine]['partition']}\n"\
            f"#SBATCH --output myjob_{self.keywords[self.engine]['name']}.out\n"\
            f"#SBATCH --error myjob_{self.keywords[self.engine]['name']}.err\n"\
            f"#SBATCH --ntasks {self.keywords[self.engine]['ntasks']}\n"\
            f"#SBATCH --cpus-per-task {self.keywords[self.engine]['cpus-per-task']}\n"\
            f"#SBATCH --time {self.keywords[self.engine]['time']}\n"\
            f"#SBATCH --job-name={self.keywords[self.engine]['name']}\n"\
            f"#SBATCH --nodes {self.keywords[self.engine]['nodes']}\n"\
            f"#SBATCH --nice={self.keywords[self.engine]['nice']}\n"\
            f"#SBATCH --gpus={self.keywords[self.engine]['gpus']}\n"
            if self.keywords[self.engine]['mail_user']:
                self.jobsh += f"#SBATCH --mail-user={self.keywords[self.engine]['mail_user']}\n"
            if self.keywords[self.engine]['exclude']:
                self.jobsh +=f"#SBATCH --exclude={self.keywords[self.engine]['exclude']}\n"

            self.jobsh +="\n# This block is for the execution of the program\n"\
            "# Setting OPENMPI paths here:\n"\
            "export PATH=\"/data/shared/opt/ORCA/openmpi314/bin:$PATH\"\n"\
            "export LD_LIBRARY_PATH=\"$LD_LIBRARY_PATH:/data/shared/opt/ORCA/openmpi314/lib\"\n\n"\
            "# Here giving the path to the ORCA binaries and giving communication protocol\n"\
            "export PATH=\"/data/shared/opt/ORCA/orca_4_2_1_linux_x86-64_openmpi314:$PATH\"\n"\
            "export LD_LIBRARY_PATH=\"/data/shared/opt/ORCA/orca_4_2_1_linux_x86-64_openmpi314:$LD_LIBRARY_PATH\"\n\n"\
            "# Creating local scratch folder for the user on the computing node.\n"\
            "MY_TEMP_DIR=\"$(mktemp -d /localdisk/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"\
            "# make sure the temp dir was created\n"\
            "[ -d $MY_TEMP_DIR ] || exit 1\n"\
            "# make sure to delete the directory on (erroneous) exit\n"\
            "trap 'rm -rf -- \"$MY_TEMP_DIR\"' EXIT\n\n"\
            "# Copy only the necessary stuff in submit directory to scratch directory. Add more here if needed.\n"\
            "cp $SLURM_SUBMIT_DIR/*.inp $MY_TEMP_DIR/\n"\
            "#cp $SLURM_SUBMIT_DIR/*.gbw $MY_TEMP_DIR/\n"\
            "#cp $SLURM_SUBMIT_DIR/*.hess $MY_TEMP_DIR/\n\n"\
            "# Creating nodefile in scratch\n"\
            "echo $SLURM_NODELIST > $MY_TEMP_DIR/${SLURM_JOB_NAME}.nodes\n\n"\
            "# This block is echoing some SLURM variables\n"\
            "echo \"Job execution start: $(date)\"\n"\
            "echo \"Jobid = $SLURM_JOBID\"\n"\
            "echo \"Host = $SLURM_JOB_NODELIST\"\n"\
            "echo \"Jobname = $SLURM_JOB_NAME\"\n"\
            "echo \"Subcwd = $SLURM_SUBMIT_DIR\"\n"\
            "echo \"SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK\"\n"\
            "echo \"SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE\"\n"\
            "echo \"The temp file used was: $MY_TEMP_DIR\"\n\n"\
            f"$(which orca) ${{SLURM_JOB_NAME}}.inp > ${{SLURM_JOB_NAME}}.log 2>&1 # run command\n\n"\
            "# ORCA has finished here. Now copy important stuff back (xyz files, GBW files etc.). Add more here if needed.\n"\
            "cp $MY_TEMP_DIR/*.gbw $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.xyz $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.opt $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.hess $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*_property.txt $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.log $SLURM_SUBMIT_DIR\n\n"\
            "#Deleting the scratch file. Normal delete on normal exit\n"\
            "rm -rf $MY_TEMP_DIR\n\n"\
            "echo \"Job execution end: $(date)\""
        elif self.engine == 'gaussian':
            # Right now we don't have access to gaussian, so, it will not be coded.
            # But is more than the same, configurate the .gjf file and the corresponded .sh
            raise Exception(f'The engine {self.engine} is not coded!')
        else:
            raise Exception(f'The engine {self.engine} is not coded!')
    def write(self, outpath, attribute):
        with open(outpath, 'w') as f:
            if attribute == 'input':
                f.write(self.input)
            elif attribute == 'jobsh':
                f.write(self.jobsh)
            else:
                raise ValueError(f"{attribute} is not a correct attribute. Must be: 'input' or 'jobsh'")

class PARAM:
    def __init__(self, machine = 'smaug', **keywords):
        self.machine = machine
        self.default_partition = {'smaug': 'deflt', 'gwdg': 'medium'}
        self.default_exclude = {'smaug': "fang[1,11-55]", 'gwdg': None}
        self.default_ntasks = {'smaug': 12, 'gwdg': 24}
        self.default_keywords = {
                #Parameters for input
                'theory': 'MP2',
                'basis': 'aug-cc-pVTZ',
                'memory': 14500,
                'name': 'mol',
                'environment': 'vacuum',
                'charge': 0,
                'coords': pd.DataFrame(),
                'min-type': None,
                'charge-type': 'ESP',
                'scan-type': 'mm',
                'dihed-fit-type': 'iterative',
                'dihed-num-iterations': 100,
                'outdir': 'outdir',
                #Parameters for jobsh
                'time': '2-00:00',
                'ntasks': self.default_ntasks[self.machine],
                'cpus-per-task': 1,
                'partition': self.default_partition[self.machine],
                'nodes': 1,
                'nice': 0,
                'gpus': 1,
                'mail_user': None,
                'exclude': self.default_exclude[self.machine]}
        self.keywords = copy.deepcopy(self.default_keywords)
        for key in keywords:
            self.keywords[key] = keywords[key]
        self.jobsh = ''
        self.parse()

    def parse(self):
        self.jobsh += "#!/bin/bash\n"\
        f"#SBATCH --partition {self.keywords['partition']}\n"\
        f"#SBATCH --output myjob_{self.keywords['name']}.out\n"\
        f"#SBATCH --error myjob_{self.keywords['name']}.err\n"\
        f"#SBATCH --ntasks {self.keywords['ntasks']}\n"\
        f"#SBATCH --cpus-per-task {self.keywords['cpus-per-task']}\n"\
        f"#SBATCH --time {self.keywords['time']}\n"\
        f"#SBATCH --job-name={self.keywords['name']}\n"\
        f"#SBATCH --nodes {self.keywords['nodes']}\n"\
        f"#SBATCH --nice={self.keywords['nice']}\n"\
        f"#SBATCH --gpus={self.keywords['gpus']}\n"
        if self.keywords['mail_user']:
            self.jobsh += f"#SBATCH --mail-user={self.keywords['mail_user']}\n"
        if self.keywords['exclude']:
            self.jobsh +=f"#SBATCH --exclude={self.keywords['exclude']}\n"

        if self.machine == 'gwdg':
            self.jobsh += f"#SBATCH -A all\n"\
            "#SBATCH -C scratch\n"

        self.jobsh +="\n# This block is for the execution of the program\n"
        if self.machine == 'smaug':
            self.jobsh += "source /home/users/all-jh/opt/miniconda3/etc/profile.d/conda.sh #It is needed to use the conda activate command\n"\
            "conda activate htmd\n\n"\
            "# Creating local scratch folder for the user on the computing node.\n"\
            "MY_TEMP_DIR=\"$(mktemp -d /localdisk/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"
        elif self.machine == 'gwdg':
            self.jobsh += "module load anaconda3/2020.07\n"\
            "source /opt/sw/rev/20.12/haswell/gcc-9.3.0/anaconda3-2020.07-slbv7z/etc/profile.d/conda.sh #It is needed to use the conda activate command\n"\
            "conda activate htmd\n\n"\
            "# Creating local scratch folder for the user on the computing node.\n"\
            "MY_TEMP_DIR=\"$(mktemp -d /scratch/users/${USER}/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"
        else:
            pass

        self.jobsh += "# make sure the temp dir was created\n"\
        "[ -d $MY_TEMP_DIR ] || exit 1\n"\
        "# make sure to delete the directory on (erroneous) exit\n"\
        "trap 'rm -rf -- \"$MY_TEMP_DIR\"' EXIT\n"\
        "# export PSI_SCRATCH env variable\n"\
        "export PSI_SCRATCH=$MY_TEMP_DIR\n\n"\
        "# This block is echoing some SLURM variables\n"\
        "echo \"Job execution start: $(date)\"\n"\
        "echo \"Jobid = $SLURM_JOBID\"\n"\
        "echo \"Host = $SLURM_JOB_NODELIST\"\n"\
        "echo \"Jobname = $SLURM_JOB_NAME\"\n"\
        "echo \"Subcwd = $SLURM_SUBMIT_DIR\"\n"\
        "echo \"SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK\"\n"\
        "echo \"SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE\"\n"\
        "echo \"The temp file used was: $MY_TEMP_DIR\"\n\n"\
        "#run paremeterize\n"\
        f"parameterize ${{SLURM_JOB_NAME}}.mol2 --memory {self.keywords['memory']} --ncpus {self.keywords['ntasks'] * self.keywords['cpus-per-task']} --theory {self.keywords['theory']} --basis {self.keywords['basis']} --environment {self.keywords['environment']} --charge {self.keywords['charge']} --min-type {self.keywords['min-type']} --charge-type {self.keywords['charge-type']} --scan-type {self.keywords['scan-type']} --dihed-fit-type {self.keywords['dihed-fit-type']} --dihed-num-iterations {self.keywords['dihed-num-iterations']} --outdir {self.keywords['outdir']} "
        "#Deleting the scratch file. Normal delete on normal exit\n"\
        "rm -rf $MY_TEMP_DIR\n\n"\
        "echo \"Job execution end: $(date)\""

    def write(self, outpath = 'job.sh'):
        with open(outpath, 'w') as f:
            f.write(self.jobsh)



class CONTINUE:
    def __init__(self, elapsed_paths, engine = 'psi4', machine = 'smaug', **keywords):
        self.elapsed_paths = elapsed_paths
        self.engine = engine
        self.machine = machine
        self.default_partition = {'smaug': 'deflt', 'gwdg': 'medium'}
        self.default_exclude = {'smaug': "fang[1,11-55]", 'gwdg': None}
        self.default_ntasks = {'smaug': 12, 'gwdg': 24}
        self.default_keywords = {
            'psi4':{
                #Parameters for jobsh
                'time': '2-00:00',
                'ntasks': self.default_ntasks[self.machine],
                'cpus-per-task': 1,
                'partition': self.default_partition[self.machine],
                'nodes': 1,
                'nice': 0,
                'gpus': 1,
                'mail_user': None,
                'exclude': self.default_exclude[self.machine],
                },
            'orca':{
                #Parameters for jobsh
                'time': '2-00:00',
                'ntasks': 6,
                'cpus-per-task': 2,
                'partition': 'deflt',
                'nodes': 1,
                'nice': 0,
                'gpus': 1,
                'mail_user': None,
                'exclude': "fang[1,11-55]" #None
                },
            'gaussian':{
                'theory': 7,
                'basis': 2,
                'memory':4,
                'name':5,
                'charge':0,
                'multiplicity':1,
                'PCM': False,
                'coords': None,
                'calc_type':5,
                'freq': False}
        }
        self.keywords = copy.deepcopy(self.default_keywords)
        for key in keywords:
            self.keywords[engine][key] = keywords[key]
        self.jobsh = ''
        self.parse()

    def parse(self):
        if self.engine == 'psi4':

            self.jobsh += "#!/bin/bash\n"\
            f"#SBATCH --partition {self.keywords[self.engine]['partition']}\n"\
            f"#SBATCH --output myjob_{self.keywords[self.engine]['name']}.out\n"\
            f"#SBATCH --error myjob_{self.keywords[self.engine]['name']}.err\n"\
            f"#SBATCH --ntasks {self.keywords[self.engine]['ntasks']}\n"\
            f"#SBATCH --cpus-per-task {self.keywords[self.engine]['cpus-per-task']}\n"\
            f"#SBATCH --time {self.keywords[self.engine]['time']}\n"\
            f"#SBATCH --job-name={self.keywords[self.engine]['name']}\n"\
            f"#SBATCH --nodes {self.keywords[self.engine]['nodes']}\n"\
            f"#SBATCH --nice={self.keywords[self.engine]['nice']}\n"\
            f"#SBATCH --gpus={self.keywords[self.engine]['gpus']}\n"
            if self.keywords[self.engine]['mail_user']:
                self.jobsh += f"#SBATCH --mail-user={self.keywords[self.engine]['mail_user']}\n"
            if self.keywords[self.engine]['exclude']:
                self.jobsh +=f"#SBATCH --exclude={self.keywords[self.engine]['exclude']}\n"

            if self.machine == 'gwdg':
                self.jobsh += f"#SBATCH -A all\n"\
                "#SBATCH -C scratch\n"

            self.jobsh +="\n# This block is for the execution of the program\n"
            if self.machine == 'smaug':
                self.jobsh += "source /home/users/all-jh/opt/miniconda3/etc/profile.d/conda.sh #It is needed to use the conda activate command\n"\
                "conda activate htmd\n\n"\
                "# Creating local scratch folder for the user on the computing node.\n"\
                "MY_TEMP_DIR=\"$(mktemp -d /localdisk/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"
            elif self.machine == 'gwdg':
                self.jobsh += "module load anaconda3/2020.07\n"\
                "source /opt/sw/rev/20.12/haswell/gcc-9.3.0/anaconda3-2020.07-slbv7z/etc/profile.d/conda.sh #It is needed to use the conda activate command\n"\
                "conda activate htmd\n\n"\
                "# Creating local scratch folder for the user on the computing node.\n"\
                "MY_TEMP_DIR=\"$(mktemp -d /scratch/users/${USER}/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"
            else:
                pass

            self.jobsh += "# make sure the temp dir was created\n"\
            "[ -d $MY_TEMP_DIR ] || exit 1\n"\
            "# make sure to delete the directory on (erroneous) exit\n"\
            "trap 'rm -rf -- \"$MY_TEMP_DIR\"' EXIT\n"\
            "# export PSI_SCRATCH env variable\n"\
            "export PSI_SCRATCH=$MY_TEMP_DIR\n\n"\
            "# This block is echoing some SLURM variables\n"\
            "echo \"Job execution start: $(date)\"\n"\
            "echo \"Jobid = $SLURM_JOBID\"\n"\
            "echo \"Host = $SLURM_JOB_NODELIST\"\n"\
            "echo \"Jobname = $SLURM_JOB_NAME\"\n"\
            "echo \"Subcwd = $SLURM_SUBMIT_DIR\"\n"\
            "echo \"SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK\"\n"\
            "echo \"SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE\"\n"\
            "echo \"The temp file used was: $MY_TEMP_DIR\"\n\n"
            for path in self.elapsed_paths:
                self.jobsh += f'cd {path} && bash run.sh && touch jobqueues.done\n'

            self.jobsh += "\n#Deleting the scratch file. Normal delete on normal exit\n"\
            "rm -rf $MY_TEMP_DIR\n\n"\
            "echo \"Job execution end: $(date)\""

        elif self.engine == 'orca':
            raise Exception(f'The engine {self.engine} is not coded!')
            # Not implemented
            self.jobsh += "#!/bin/bash\n"\
            f"#SBATCH --partition {self.keywords[self.engine]['partition']}\n"\
            f"#SBATCH --output myjob_{self.keywords[self.engine]['name']}.out\n"\
            f"#SBATCH --error myjob_{self.keywords[self.engine]['name']}.err\n"\
            f"#SBATCH --ntasks {self.keywords[self.engine]['ntasks']}\n"\
            f"#SBATCH --cpus-per-task {self.keywords[self.engine]['cpus-per-task']}\n"\
            f"#SBATCH --time {self.keywords[self.engine]['time']}\n"\
            f"#SBATCH --job-name={self.keywords[self.engine]['name']}\n"\
            f"#SBATCH --nodes {self.keywords[self.engine]['nodes']}\n"\
            f"#SBATCH --nice={self.keywords[self.engine]['nice']}\n"\
            f"#SBATCH --gpus={self.keywords[self.engine]['gpus']}\n"
            if self.keywords[self.engine]['mail_user']:
                self.jobsh += f"#SBATCH --mail-user={self.keywords[self.engine]['mail_user']}\n"
            if self.keywords[self.engine]['exclude']:
                self.jobsh +=f"#SBATCH --exclude={self.keywords[self.engine]['exclude']}\n"

            self.jobsh +="\n# This block is for the execution of the program\n"\
            "# Setting OPENMPI paths here:\n"\
            "export PATH=\"/data/shared/opt/ORCA/openmpi314/bin:$PATH\"\n"\
            "export LD_LIBRARY_PATH=\"$LD_LIBRARY_PATH:/data/shared/opt/ORCA/openmpi314/lib\"\n\n"\
            "# Here giving the path to the ORCA binaries and giving communication protocol\n"\
            "export PATH=\"/data/shared/opt/ORCA/orca_4_2_1_linux_x86-64_openmpi314:$PATH\"\n"\
            "export LD_LIBRARY_PATH=\"/data/shared/opt/ORCA/orca_4_2_1_linux_x86-64_openmpi314:$LD_LIBRARY_PATH\"\n\n"\
            "# Creating local scratch folder for the user on the computing node.\n"\
            "MY_TEMP_DIR=\"$(mktemp -d /localdisk/psi4_${SLURM_JOBID}_$(date +%Y-%m-%d-%H-%M-%S)-XXXXXXXXXX)\"\n"\
            "# make sure the temp dir was created\n"\
            "[ -d $MY_TEMP_DIR ] || exit 1\n"\
            "# make sure to delete the directory on (erroneous) exit\n"\
            "trap 'rm -rf -- \"$MY_TEMP_DIR\"' EXIT\n\n"\
            "# Copy only the necessary stuff in submit directory to scratch directory. Add more here if needed.\n"\
            "cp $SLURM_SUBMIT_DIR/*.inp $MY_TEMP_DIR/\n"\
            "#cp $SLURM_SUBMIT_DIR/*.gbw $MY_TEMP_DIR/\n"\
            "#cp $SLURM_SUBMIT_DIR/*.hess $MY_TEMP_DIR/\n\n"\
            "# Creating nodefile in scratch\n"\
            "echo $SLURM_NODELIST > $MY_TEMP_DIR/${SLURM_JOB_NAME}.nodes\n\n"\
            "# This block is echoing some SLURM variables\n"\
            "echo \"Job execution start: $(date)\"\n"\
            "echo \"Jobid = $SLURM_JOBID\"\n"\
            "echo \"Host = $SLURM_JOB_NODELIST\"\n"\
            "echo \"Jobname = $SLURM_JOB_NAME\"\n"\
            "echo \"Subcwd = $SLURM_SUBMIT_DIR\"\n"\
            "echo \"SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK\"\n"\
            "echo \"SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE\"\n"\
            "echo \"The temp file used was: $MY_TEMP_DIR\"\n\n"\
            f"$(which orca) ${{SLURM_JOB_NAME}}.inp > ${{SLURM_JOB_NAME}}.log 2>&1 # run command\n\n"\
            "# ORCA has finished here. Now copy important stuff back (xyz files, GBW files etc.). Add more here if needed.\n"\
            "cp $MY_TEMP_DIR/*.gbw $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.xyz $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.opt $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.hess $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*_property.txt $SLURM_SUBMIT_DIR\n"\
            "cp $MY_TEMP_DIR/*.log $SLURM_SUBMIT_DIR\n\n"\
            "#Deleting the scratch file. Normal delete on normal exit\n"\
            "rm -rf $MY_TEMP_DIR\n\n"\
            "echo \"Job execution end: $(date)\""
        elif self.engine == 'gaussian':
            # Right now we don't have access to gaussian, so, it will not be coded.
            # But is more than the same, configurate the .gjf file and the corresponded .sh
            raise Exception(f'The engine {self.engine} is not coded!')
        else:
            raise Exception(f'The engine {self.engine} is not coded!')
    def write(self, outpath):
        with open(outpath, 'w') as f:
            f.write(self.jobsh)



if __name__ == '__main__':

    s = CONTINUE(['./68', 'klusef/fsdg/oi/'],machine = 'gwdg', name = 'elapse1')
    #s = INPUT('psi4',machine = 'gwdg', name =  'imi')
    print(s.jobsh)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import rmsd
import numpy as np
import pandas as pd
import os
from aleimi import utils

def arc_reader (arc:str):
    """This , for sure, will change in the future to a more object-oriented paradigm
    It reads a arc MOPAC file and give a tuple of

    Parameters
    ----------
    arc : str
        The arc file path

    Returns
    -------
    list[tuple]
        A list of tuple sorted by energy. Each tuple contains:
            #. cells: conformer identifier,
            #. HeatsOfFormation_kcalmol: self-explanatory,
            #. CONTAINER__H: Coordinates numpy array of the heavy atoms with shape (Number of heavy atoms, 3),
            #. Class_E: The classic energy calculated during :meth:`aleimi.confgen.main`
    """
    with open(arc, 'rt', encoding='latin-1') as a:
        lines = a.readlines()

    # getting data from arc
    HeatsOfFormation_kcalmol = []
    cells = []
    Class_E = []
    # finding No. of atoms
    for line in lines:
        if 'Empirical Formula' in line:
            natoms = int(line.split()[-2])
            break
    
    #CONTAINER = []
    CONTAINER__H = []
    #cart = []
    cart__H = []
    #atoms = []
    for i, line in enumerate(lines):

        if 'HEAT OF FORMATION' in line:
            #I am getting the one in kcal/mol
            HeatsOfFormation_kcalmol.append(float(line.split()[-5])) #HEAT OF FORMATION       =       -180.69875 KCAL/MOL =    -756.04358 KJ/MOL
       
        elif 'Empirical Formula' in line:
            try:
                Class_E.append(float(lines[i+3].split('=')[-1].split()[0])) #las dos versiones con y sin optimizacion
            except:
                Class_E.append(None)
                    
            cells.append(int(lines[i+4].split(':')[1]))
    
        elif ('FINAL GEOMETRY OBTAINED' in line):
            chunk = lines[i+4:i+4+natoms]
            cart__H = []
            for c in chunk:
                #atoms.append(c.split()[0])
                #cart.append([c.split()[0].strip(), float(c.split()[1]), float(c.split()[3]), float(c.split()[5])])
                if c.split()[0] != 'H':
                    cart__H.append([c.split()[1], c.split()[3], c.split()[5]]) # No estoy tomando los atomos, solamente coordenadas c.split()[0].strip(),
            #CONTAINER.append(np.asarray(pd.DataFrame(cart)))
            CONTAINER__H.append(np.array(cart__H, dtype=np.float64))
    #atoms = (np.asarray(atoms))
    # .... organizing
    paired = list(zip(cells, HeatsOfFormation_kcalmol, CONTAINER__H, Class_E)) # Esto genera un arreglo de tuplas, me une los arreglos
    ORDERED = sorted(paired, key=lambda x: x[1])  #Esto ordena la tupla segun la energia de menor a mayor
    return ORDERED #, atoms]

def out_reader (out):
    """This , for sure, will change in the future to a more object-oriented paradigm
    It reads a out MOPAC file and give a tuple of

    Parameters
    ----------
    out : str
        The arc file path

    Returns
    -------
    list[tuple]
        A list of tuple sorted by energy. Each tuple contains:
            #. cells: conformer identifier,
            #. HeatsOfFormation_kcalmol: self-explanatory,
            #. CONTAINER__H: Coordinates numpy array of the heavy atoms with shape (Number of heavy atoms, 3)
            #. Class_E: The classic energy calculated during :meth:`aleimi.confgen.main`
    """
    f = open(out, 'r')
    chunk = []
    HeatsOfFormation_kcalmol = []
    cells = []
    Class_E = []
    CONTAINER__H = []
    cart__H = []


    # getting data from out                                                       #
    # finding No. of atoms
    while True:
        line = f.readline()
        if "Empirical Formula" in line:
            natoms = int(line.split()[-2])
            break

    f = open(out, 'r')
    while True:
        line = f.readline()
        if len(line) == 0:break
               
        if 79*'-' in line:
            while True:
                line = f.readline()
                if (79*'*' in line) or (len(line) == 0):break
                if "E_UFF = " in line:
                    split = line.split('=')[-1]
                    try:
                        Class_E.append(float(split))
                    except:
                        Class_E.append(None)

                elif 'CELL' in line:
                    cells.append(int(line.split(':')[1]))
                    
                elif 'HEAT OF FORMATION' in line:
                    HeatsOfFormation_kcalmol.append(float(line.split()[-5]))
                elif 'CARTESIAN COORDINATES' in line:
                    utils.ignoreLines(f, 1)
                    cont = 0
                    chunk = []
                    while cont < natoms:
                        chunk.append(f.readline())
                        cont += 1
                    cart__H = []
                    for c in chunk:
                        if c.split()[1] != 'H':
                            cart__H.append([c.split()[2], c.split()[3], c.split()[4]]) # No estoy tomando los atomos, solamente coordenadas c.split()[0].strip(),

            CONTAINER__H.append(np.array(cart__H, dtype=np.float64))
    f.close
    # .... organizing
    paired = list(zip(cells, HeatsOfFormation_kcalmol, CONTAINER__H, Class_E)) # Esto genera un arreglo de tuplas, me une los arreglos
    ORDERED = sorted(paired, key=lambda x: x[1])  #Esto ordena la tupla segun la energia de menor a mayor
    return ORDERED

def main(file_path:str, Bd_rmsd:float = 1.0, Bd_E:float = 0.0, BOutPath:bool = True) -> pd.DataFrame:
    """It reads the MOPAC output file (arc or out) and create a Boltzmann table

    Parameters
    ----------
    file_path : str
        MOPAC output (.arc or .out). It will be read with :meth:`aleimi.boltzmann.arc_reader` or :meth:`aleimi.boltzmann.out_reader`
        depending on the extension.
    Bd_rmsd : float, optional
        RMSD to filter out redundant conformations, geometric filter, by default 1.0
    Bd_E : float, optional
        Energy difference in kJ to filter out redundant conformations, geometric filter, by default 1.0, by default 0.0
    BOutPath : bool, optional
        Directory to ouput the table {file_name}_boltzmann.csv", by default True

    Returns
    -------
    pd.DataFrame

    A Table with columns:
        #. `cell`: conformer identifier,
        #. `Class_E`: Classic energy from the RDKit optimization. in this example is `NaN` because we did not perform this optimization,
        #. `HeatOfFormation_kcal`: self-explanatory [kcal/mol],
        #. `Emin_Ei`: Difference in energy between the lower and the i-th conformer in [kcal/mol],
        #. `qi__Pi/Pmin__e^(Emin_Ei)/KbT`: Boltzmann factors,
        #. `Fraction_%__100*qi/q`: Occupancy of each conformer,

    Raises
    ------
    ValueError
        _description_
    """
    name, ext = os.path.splitext(os.path.basename(file_path))
    
    if ext == '.arc':
        ordered = (arc_reader(file_path))
    elif ext == '.out':
        ordered = (out_reader(file_path))
    else:
        raise ValueError(f"{file_path} does not have .arc or .out extension. Therefore is not readable by ALEIMI.")

    if Bd_E:
        for i, _ in enumerate(ordered):
            to_trash_degenerated = []
            for idx, _ in enumerate(ordered):
                if i < idx:
                    Ei = ordered[i][1]
                    Eidx = ordered[idx][1]
                    delta = abs(Ei - Eidx)
                    if delta <= Bd_E:
    # =============================================================
    #     CHECKING Geometric degeneracy
    # =============================================================
                        P = ordered[i][2]
                        Q = ordered[idx][2]

                        RMSD = rmsd.kabsch_rmsd(P, Q, translate = True)

                        if RMSD <= Bd_rmsd:
                        # reject identical structure
                            to_trash_degenerated.append(idx)
# =========================================================================
#     FOR EACH STRUCTURE, eliminate degenerated and save lot of time
# =========================================================================
            to_trash_degenerated = sorted(to_trash_degenerated, reverse=True)
            _ = [ordered.pop(x) for x in to_trash_degenerated]

    else:
        for i, x in enumerate(range(len(ordered))):
            to_trash_degenerated = []
            for idx, y in enumerate(range(len(ordered))):
                if i < idx:
                
# =============================================================
#     CHECKING Geometric degeneracy
# =============================================================
                    P = ordered[i][2]
                    Q = ordered[idx][2]

                    RMSD = rmsd.kabsch_rmsd(P, Q, translate = True)

                    if RMSD <= Bd_rmsd:
                        # reject identical structure and kept the lowest energy (because ordered() is ordered using the enrgy, so idx always will have a grater enrgy)
                        to_trash_degenerated.append(idx)

# =========================================================================
#     FOR EACH STRUCTURE, eliminate degenerated and save lot of time
# =========================================================================
            to_trash_degenerated = sorted(to_trash_degenerated, reverse=True)
            [ordered.pop(x) for x in to_trash_degenerated]
        


# =============================================================================
#      WORKING with UNDEGENERATED. Cambie la manera de calculos los parametros:
#Me base en: James B. Foresman - Exploring Chemistry With Electronic Structure Methods 3rd edition (2015) pag 182
# y Mortimer_Physical Chemistry_(3rd.ed.-2008) pag 1045
# =============================================================================
    Kb = 1.987204259E-3                        # kcal/(mol⋅K)
    T = 298.15                                 # Absolute T (K)
    DF = pd.DataFrame()
    cells = [ordered[i][0] for i, x in enumerate(ordered)]
    Class_E = [ordered[i][3] for i, x in enumerate(ordered)]
    HeatsOfFormation_kcalmol = [ordered[i][1] for i, x in enumerate(ordered)]
    MinHeatsOfFormation_kcalmol = min(HeatsOfFormation_kcalmol)
    relative_kcalmol = [MinHeatsOfFormation_kcalmol - x for x in HeatsOfFormation_kcalmol]
    qi = [np.exp(E_r/(Kb*T)) for E_r in relative_kcalmol]
    q = sum(qi)
    Fraction = [100*i/q for i in qi]
    #Z = [np.e**(-(E/(k*T))) for E in energy_kcal] #no pudo calcular Z: verflowError: (34, 'Result too large')
    #Pi_b = [(np.e**-(E/(k*T)))/Z for E in energy_kcal]
    # =============================================================================
    #     DATAFRAME
    # =============================================================================
    DF['cell'] = cells
    DF['Class_E'] = Class_E
    DF['HeatOfFormation_kcal/mol'] = HeatsOfFormation_kcalmol
    DF['Emin_Ei'] = relative_kcalmol
    DF['qi__Pi/Pmin__e^(Emin_Ei)/KbT'] = qi
    DF['Fraction_%__100*qi/q'] = Fraction

    if BOutPath:
        with open(f"{name}_boltzmann.csv", 'wt') as rp:
            DF.to_csv(rp)
    
    return DF


if __name__ == '__main__':...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os, subprocess, time, inspect

def timeit(method):
    """Useful as decorator
    """
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % \
                  (method.__name__, (te - ts) * 1000))
        return result
    return timed

        

def run(command: str, shell: bool = True, executable: str = '/bin/bash'):
    """This function is just a useful wrapper around subprocess.run

    Parameters
    ----------
    command : str
        Any command to execute.
    shell : bool, optional
        keyword of ``subprocess.Popen`` and ``subprocess.Popen``, by default True
    executable : str, optional
        keyword of ``subprocess.Popen`` and ``subprocess.Popen``, by default '/bin/bash'

    Returns
    -------
    object
        The processes returned by Run.

    Raises
    ------
    RuntimeError
        In case of non-zero exit status on the provided command.
    """

    process = subprocess.run(command, shell=shell, executable=executable, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    returncode = process.returncode
    if returncode != 0:
        # print(f'Command {command} returned non-zero exit status {returncode}')
        raise RuntimeError(process.stderr)
    return process

@timeit
def mopac(mop:str):
    """A simple wrapper around MOPAC

    Parameters
    ----------
    mop : str
        The path where the .mop file is located
    """
    print(f"Mopac is running ...")
    run(f"mopac {mop}  > /dev/null 2>&1")
    print("Done!")

def makedirs(path):
    os.makedirs(path,exist_ok=True)

def get_default_kwargs(func:object) -> dict:
    """Get the keywords arguments of a function

    Parameters
    ----------
    func : object
        The function to analysis

    Returns
    -------
    dict
        keyword:keyword_value
    """
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

def ignoreLines(file_obj:object, num_of_lines_to_skip:int):
    """Ignore lines during reading of file

    Parameters
    ----------
    file_obj : object
        The file in open mode
    num_of_lines_to_skip : int
        Number of lines to skip
    """
    for _ in range(num_of_lines_to_skip): file_obj.readline()


if __name__ == '__main__':...

  